import bpy
import os
import bmesh
from mathutils import Vector, Color
from bpy.props import FloatProperty, BoolProperty, EnumProperty
import random


bl_info = {
    "name": "DJH Utils",
    "description": "Useful operators that I have wanted at one time or another.",
    "author": "Daniel Hickox",
    "version": (1, 43, 0), 
    "blender": (3, 5, 0), # Min 3.5 Updated for 4.1
    "category": "3D View",
    "doc_url": "https://danielhickox.com/utils",
}


classes = []

class DJH_OT_GetMatTextures(bpy.types.Operator):
    """Searches for textures with the same base name and adds them to the material."""
    bl_idname = "djh.get_textures"
    bl_label = "Get Material Textures"
    bl_options = {'REGISTER', 'UNDO'}

    scale = None

    skip_unused: BoolProperty(name="Skip Unused Images")

    def execute(self, context):
        # used to convert possible texture names to socket types
        lookup = {
            "D": ["basecolor", "diffuse", "col", "albedo", "color"],
            "A": ["ao"],
            "M": ["metallic", "metal"],
            "R": ["roughness", "rough"],
            "N": ["normal", "nor"],
            "B": ["bump", "height"],
        }

        # returns the last index of c in the path
        def idx(c, path):
            x = 0
            y = 0
            while x != -1:
                y = x
                x = path.find(c, x+1)
            return y
            
        # returns the corresponding type from the lookup dictionary
        def lookup_type(attr):            
            for atype in lookup:
                if attr.lower() in lookup[atype]:
                    return atype

        # used to check if a node should be created
        def test_connection(atype, name, shade):
            if not self.skip_unused: return True
            for input in shade.inputs:
                if atype == None and input.name.lower() == name.lower():
                    return True
                elif atype != None and input.name.lower() in lookup[atype]:
                    return True
            return False

        # links the image output to the correct input socket of the shader node based on the attribute type
        def make_link(tree, atype, img, shade, name = ""):      
            for input in shade.inputs:
                if atype == None and input.name.lower() == name.lower():
                    tree.links.new(img.outputs[0], input)
                    break     
                elif atype != None and input.name.lower() in lookup[atype]:
                    tree.links.new(img.outputs[0], input)
                    break     
        
        def normal_type(shade):      
            for input in shade.inputs:
                if input.name.lower() in lookup["N"]:
                    return input.type

        # creates a new image texture node with the given name
        def make_node(name, origin, file, idx):
            tree = origin.id_data
            shader = origin

            # If the scale node doesn't exist yet, create it and connect it to the shader tree
            if self.scale == None:
                uv = tree.nodes.new('ShaderNodeUVMap')
                uv.location = origin.location + Vector([-600, 0])
                self.scale = tree.nodes.new('ShaderNodeVectorMath')
                self.scale.location = origin.location + Vector([-300, 0])
                self.scale.operation = 'SCALE'
                self.scale.label = 'UV Scale'
                tree.links.new(uv.outputs['UV'], self.scale.inputs[0])
                tree.links.new(self.scale.outputs['Vector'], origin.inputs['Vector'])

            # Assumes the shader node is the input node connected to output node
            for n in tree.nodes:
                print(n.bl_idname)
                if(n.bl_idname == "ShaderNodeOutputMaterial"):
                    shader = n.inputs[0].links[0].from_node
                    break
                
            # Look up the type of the texture
            atype = lookup_type(name)

            if not test_connection(atype, name, shader): return False

            # Create a new image texture node
            node = tree.nodes.new('ShaderNodeTexImage')
            node.label = name
            node.location = origin.location + Vector([0, -300 * idx])


            # Try to load the image file
            try:
                img = bpy.data.images.load(file)
                
                if atype != 'C' and atype != None:
                    img.colorspace_settings.name = 'Non-Color'

                node.image = img
            except:
                print('no file')
  
            # Additional nodes for special types
            if atype == 'N' and (normal_type(shader) == 'VECTOR'):
                nm = tree.nodes.new('ShaderNodeNormalMap')
                nm.location = node.location + Vector([300, 0])
                tree.links.new(node.outputs['Color'], nm.inputs['Color'])
                make_link(tree, atype, nm, shader)
            elif atype == 'B':
                bm = tree.nodes.new('ShaderNodeBump')
                bm.location = node.location + Vector([300, 0])
                tree.links.new(node.outputs['Color'], bm.inputs['Height'])
                make_link(tree, atype, bm, shader)
            else:
                make_link(tree, atype, node, shader, name)

            # Connect the scale node to the texture node
            tree.links.new(self.scale.outputs['Vector'], node.inputs['Vector'])
            return True

        node = context.active_node
        if node.bl_idname == 'ShaderNodeTexImage':
            img = node.image
            path = bpy.path.abspath(img.filepath)

            # Split the path into parts
            i = idx('.', path)
            ext = path[i:]
            path = path[:i]
            i = idx('\\', path)
            name = path[i+1:]
            path = path[:i+1]
            i = idx('_', name)
            attr = name[i+1:]
            name = name[:i]
            print(path, name, attr, ext)

            node.label = attr

            maps = []
            
            # Loop through the files in the same folder and keep any that start with the same base name
            for file in os.listdir(path):
                if file.startswith(name):
                    maps.append(file)

            print(maps)
            
            # Allowed types
            extensions = ['.png', '.tga', '.jpg', '.tiff', '.exr', '.PNG', '.TGA', '.JPG', '.TIFF', '.EXR']

            #Get Scale Node or reset to None
            self.scale = None
            for n in node.id_data.nodes:
                if(n.label == "UV Scale"):
                    self.scale = n
            
            id = 1
            # Loop through the found files and make a node for each valid texture
            for file in maps:
                i = idx('.', file)
                if file[i:] not in extensions: continue
                fn = file[:i]
                i = idx('_', fn)
                a = fn[i+1:]
                if a == attr: continue

                n = fn[:i]
                print(n, a, fn)

                if make_node(a, node, f'{path}{file}', id):
                    id+=1

        return {'FINISHED'}
    
classes.append(DJH_OT_GetMatTextures)


class DJH_OT_NodeLinks(bpy.types.Operator):
    """Creates many links between two nodes"""
    bl_idname = 'djh.bridge_nodes'
    bl_label = 'Bridge Node Sockets (by name)'
    bl_options = {'REGISTER', 'UNDO'}

    @classmethod
    def poll(cls, context):
        space = context.space_data
        valid_trees = ["ShaderNodeTree", "CompositorNodeTree", "TextureNodeTree", "GeometryNodeTree"]

        valid = False
        if space.type == 'NODE_EDITOR' and space.node_tree is not None and space.tree_type in valid_trees:
            valid = True

        return valid

    def execute(self, context):
        # the current node tree
        tree = context.space_data.node_tree 

        sel = []
        # build a list of the selected nodes
        for n in tree.nodes:
            if n.select:
                sel.append(n)

        # don't try if more/less than two nodes are selected
        if len(sel) != 2:
            self.report({'INFO'}, "Wrong number of nodes selected")
            return {'CANCEL'}

        # decide which is node comes first in the change based on X position
        n1 = sel[0]
        n2 = sel[1]
        if sel[0].location.x > sel[1].location.x:
            n1 = sel[1]
            n2 = sel[0]

        # make the links
        for o in n1.outputs:
            for i in n2.inputs:
                if o.name == i.name:
                    tree.links.new(o, i)

        return {'FINISHED'}
    
classes.append(DJH_OT_NodeLinks)


class DJH_OT_CleanWeights(bpy.types.Operator):
    """Removes verts with zero weight from groups then removes groups with zero verts from mesh"""
    bl_idname = 'djh.clean_vert_groups'
    bl_label = 'Clean Vertex Groups'
    bl_options = {'REGISTER', 'UNDO'}

    def execute(self, context):
        o = context.active_object

        list = []

        # remove verts from the group if their value == 0
        for v in o.data.vertices:
            for g in v.groups:
                if g.weight == 0:
                    continue
                if g.group not in list:
                    list.append(g.group)

        # remove groups from the mesh if they don't have any verts assigned
        for i in range(len(o.vertex_groups)-1, -1, -1):
            if o.vertex_groups[i].index not in list:
                o.vertex_groups.remove(o.vertex_groups[i])

        return {'FINISHED'}

classes.append(DJH_OT_CleanWeights)


class DJH_OT_ModAutoName(bpy.types.Operator):
    """Renames modifiers on the selected objects to reflect the modifier type or node_group"""
    bl_idname = 'djh.rename_modifiers'
    bl_label = 'Auto-Name Modifiers'
    bl_options = {'REGISTER', 'UNDO'}

    # make modifier labels match their type or node group name
    def execute(self, context):
        for o in context.selected_objects:
            for m in o.modifiers:
                m.name = m.type
                if m.type == 'NODES':
                    m.name = m.node_group.name
        return {'FINISHED'}

classes.append(DJH_OT_ModAutoName)


class DJH_OT_Snap_To_Faces(bpy.types.Operator):
    """Snaps the selected geometry to the nearest faces"""
    bl_idname = 'djh.snap_to_faces'
    bl_label = 'Snap to Near Faces'
    bl_options = {'REGISTER', 'UNDO'}

    def execute(self, context):
        bpy.ops.transform.translate(value=(0, 0, 0), use_proportional_edit=False, snap=True, snap_elements={'FACE_NEAREST'}, use_snap_self=False, use_snap_edit=False, use_snap_nonedit=True)
        return {'FINISHED'}

classes.append(DJH_OT_Snap_To_Faces)

class DJH_OT_Instance_Materials(bpy.types.Operator):
    """Creates placeholder materials on the active object from the given collection"""
    bl_idname = 'djh.placeholder_materials'
    bl_label = 'Create Materials for Instances'
    bl_options = {'REGISTER', 'UNDO'}

    
    def get_collections(self, context):
        collections = bpy.data.collections
        collection_names = [(col.name, col.name, '') for col in collections]
        return collection_names
    
    # instances: StringProperty( name="Instance Collection" )
    collection: bpy.props.EnumProperty(items=get_collections, name="Instances Collection")
    
    def execute(self, context):
        print("Selected Collection:", self.collection)

        col = bpy.data.collections[self.collection]
        o = context.object
        o.data.materials.clear()

        keys = col.objects.keys()
        keys.sort()

        for key in keys:
            name = f"idx_{key}"                
            if name not in bpy.data.materials:
                m = bpy.data.materials.new(name)
                m.use_nodes = True
                bsdf = m.node_tree.nodes[0]
                print(bsdf.name)
                wf = m.node_tree.nodes.new("ShaderNodeWireframe")
                m.node_tree.links.new(wf.outputs[0], bsdf.inputs[4])
                wf.use_pixel_size = True
                wf.inputs[0].default_value = 5

                c = Color()
                c.hsv = (random.random(), 0.9, 0.8)
                bsdf.inputs[0].default_value = (c.r, c.g, c.b, 1)
                m.blend_method = 'CLIP'
                m.shadow_method = 'NONE'

            o.data.materials.append(bpy.data.materials[name])

        return {'FINISHED'}

    def invoke(self, context, event):
        wm = context.window_manager
        return wm.invoke_props_dialog(self)

classes.append(DJH_OT_Instance_Materials)
    
    
    
class DJH_OT_Randomize_Seeds(bpy.types.Operator):
    """Looks for properties named 'seed' in the selected objects and sets them a to random value"""
    bl_idname = 'djh.randomize_seeds'
    bl_label = 'Randomize Seeds'
    bl_options = {'REGISTER', 'UNDO'}

    def execute(self, context):        
        def TryRandomize(o, m, name):
            if bpy.app.version[0] >= 4:
                if m.node_group.interface.items_tree.get(name) != None:
                    m[m.node_group.interface.items_tree[name].identifier] = random.randint(0, 9999999)
                    o.data.update_tag()
            else:
                if m.node_group.inputs.get(name) != None:
                    m[m.node_group.inputs[name].identifier] = random.randint(0, 9999999)
                    o.data.update()

        for o in context.selected_objects:
            for m in o.modifiers:
                if m.type == 'NODES':
                    TryRandomize(o, m, 'Seed')
                    TryRandomize(o, m, 'seed')

        if bpy.app.version[0] >= 4:
            context.area.tag_redraw()
            context.view_layer.update()

        return {'FINISHED'}

classes.append(DJH_OT_Randomize_Seeds)

    
class DJH_OT_KnifeProjectSelected(bpy.types.Operator):
    """Does a knife projection with the selected geometry in edit mode."""
    bl_idname = 'djh.knife_project_sel'
    bl_label = 'Knife Project Selection'
    bl_options = {'REGISTER', 'UNDO'}

    delete: BoolProperty(name="Delete Cutter Geo", default=True)
    merge: BoolProperty(name="Merge Verts", default=True)
    cut_through: BoolProperty(name="Cut Through", default=True)

    @classmethod
    def poll(self, context):
        if not context.mode == 'EDIT_MESH':
            self.poll_message_set("Must be in edit-mode to use")
            return False

        if len(context.selected_objects) > 1:
            self.poll_message_set("Cannot be used when editing multiple meshes")
            return False
        
        return True

    def execute(self, context):
        obj = context.object
        m : bpy.types.Mesh = context.object.data
        
        # Make a copy of the edit mesh to use as the knife
        bpy.ops.object.mode_set(mode='OBJECT')
        cutter = bpy.data.objects.new("djh_knife_pro_cutter", m)
        cutter.matrix_world = context.object.matrix_world
        context.scene.collection.objects.link(cutter)
        cutter.data = m

        # Make a copy of the mesh data so we can delete part of it
        cutter_mesh = bpy.data.meshes.new_from_object(cutter)
        cutter_mesh.name = "djh_knife_pro_cutter_mesh"
        cutter.data = cutter_mesh

        # Delete selection from original object
        if self.delete:
            bpy.ops.object.mode_set(mode='EDIT')
            bm = bmesh.from_edit_mesh(m)
            verts = [v for v in bm.verts if v.select]
            bmesh.ops.delete(bm, geom=verts)            
            bpy.ops.object.mode_set(mode='OBJECT')

        #Delete geo that is not selected from cutter object
        bpy.ops.object.select_all(action='DESELECT')
        cutter.select_set(True)
        bpy.ops.object.mode_set(mode='EDIT')
        bm = bmesh.from_edit_mesh(cutter_mesh)
        verts = [v for v in bm.verts if not v.select]
        bmesh.ops.delete(bm, geom=verts)
        bmesh.ops.delete(bm, geom=bm.faces, context='FACES_ONLY')

        # Remember how many verts are in the mesh
        vcount = len(m.vertices)

        # Make the cut
        bpy.ops.object.mode_set(mode='OBJECT')
        cutter.select_set(False)
        obj.select_set(True)
        bpy.context.view_layer.objects.active = obj
        bpy.ops.object.mode_set(mode='EDIT')
        cutter.select_set(True)

        if bpy.app.version[0] >= 4:
            bpy.ops.mesh.knife_project(cut_through=self.cut_through)
        else:
            bpy.ops.mesh.knife_project(context.copy(), cut_through=self.cut_through)
            
        bpy.ops.mesh.select_mode(type='VERT')
        bpy.ops.mesh.select_all(action='DESELECT')
        bpy.ops.object.mode_set(mode='OBJECT')

        # Remove the cutter
        context.scene.collection.objects.unlink(cutter)
        bpy.data.objects.remove(cutter)
        bpy.data.meshes.remove(cutter_mesh)

        # Select the new verts
        for i in range(vcount, len(m.vertices)):
            m.vertices[i].select = True            
        bpy.ops.object.mode_set(mode='EDIT')

        if self.merge:
            bpy.ops.mesh.remove_doubles(use_unselected=False)

        return {'FINISHED'}

classes.append(DJH_OT_KnifeProjectSelected)


    
class DJH_OT_SaveShapeAndFreeze(bpy.types.Operator):
    """Applies the active modifier as a shape and then removes the shapes keys"""
    bl_idname = 'djh.freeze_modifier'
    bl_label = 'Freeze Modifier into Mesh'
    bl_options = {'REGISTER', 'UNDO'}

    @classmethod
    def poll(self, context):
        if context.object.type != 'MESH':
            self.poll_message_set("Can only be used on a mesh object")
            return False
        
        if context.object.data.shape_keys == None:
            return True

        if len(context.object.data.shape_keys) > 0:
            self.poll_message_set("The mesh already has shape keys, please remove them to avoid data loss")
            return False
        
        if context.object.modifiers.active == None:
            self.poll_message_set("No modifier is selected")
            return False
        
        return True

    def execute(self, context):
        obj = context.object
        data = obj.data

        bpy.ops.object.modifier_apply_as_shapekey(keep_modifier=True, modifier= obj.modifiers.active.name)

        obj.active_shape_key_index = 0
        bpy.ops.object.shape_key_remove()
        bpy.ops.object.shape_key_remove()

        return {'FINISHED'}

classes.append(DJH_OT_SaveShapeAndFreeze)


    
class DJH_OT_ModifierApplyAbove(bpy.types.Operator):
    """Applies all modifiers above and including the active modifier, even if the object is a curve"""
    bl_idname = 'djh.modifier_apply_above'
    bl_label = 'Apply Modifiers Above'
    bl_options = {'REGISTER', 'UNDO'}

    data_type: EnumProperty(
        name="Data Type", 
        default='AUTO',
        description="The default format to use when saving assets",
        items=[
            ('AUTO',"Auto","Tries to keep the original type, unless it was a curve and now has faces."),
            ('MESH',"Mesh","The applied object will be a mesh."),
            ('CURVE',"Curve","The applied object will be a curve.")
        ])

    @classmethod
    def poll(self, context):
        if context.object.type not in ['MESH', 'CURVE']:
            self.poll_message_set("Can only be used on a meshes and curves")
            return False
                
        if context.object.modifiers.active == None:
            self.poll_message_set("No modifier is selected")
            return False
        
        return True

    def execute(self, context):
        o = context.object

        #rename source object so current name is available
        obj_name = o.name
        data_name = o.data.name

        o.name = f'{o.name}_source'
        o.data.name = f'{o.data.name}_source'

        copy_list = []

        apply = True
        for m in o.modifiers:
            if not apply:
                copy_list.append((m.show_viewport, m))
                m.show_viewport = False

            if m == o.modifiers.active:
                apply = False            

        dg = context.evaluated_depsgraph_get()
        data = bpy.data.meshes.new_from_object(o.evaluated_get(dg), depsgraph=dg)
        data.name = data_name
        obj = bpy.data.objects.new(obj_name, data)
        obj.matrix_world = o.matrix_world
        o.users_collection[0].objects.link(obj)

        for m in o.modifiers:
            if m.show_viewport:
                o.modifiers.remove(m)

        for m in copy_list:
            m[1].show_viewport = m[0]

        if (self.data_type == 'AUTO' and o.data.id_type == 'CURVE' and len(obj.data.polygons) == 0) or self.data_type == 'CURVE':
            bpy.ops.object.select_all(action='DESELECT')
            obj.select_set(True)
            context.view_layer.objects.active = obj
            bpy.ops.object.convert(target='CURVE')

        bpy.ops.object.select_all(action='DESELECT')
        o.select_set(True)
        obj.select_set(True)
        context.view_layer.objects.active = o
        bpy.ops.object.make_links_data(type='MODIFIERS')

        if o.data.id_type == 'CURVE':
            bpy.data.curves.remove(o.data)
        else:
            bpy.data.meshes.remove(o.data)

        context.view_layer.objects.active = obj

        return {'FINISHED'}

classes.append(DJH_OT_ModifierApplyAbove)

class DJH_OT_NodesSelectByLoc(bpy.types.Operator):
    """Selects all nodes to the left or right of the cursor."""
    bl_idname = 'djh.select_nodes_by_loc'
    bl_label = 'Select Nodes by Location'
    bl_options = {'REGISTER', 'UNDO'}

    location: FloatProperty(name="Location", default=0)
    direction: EnumProperty(
        name="Direction", 
        default='LEFT',
        description="The direction to select in",
        items=[
            ('LEFT',"Left","Select left"),
            ('RIGHT',"Right","Select Right"),
        ])

    @classmethod
    def poll(cls, context):
        space = context.space_data
        valid_trees = ["ShaderNodeTree", "CompositorNodeTree", "TextureNodeTree", "GeometryNodeTree"]

        valid = False
        if space.type == 'NODE_EDITOR' and space.node_tree is not None and space.tree_type in valid_trees:
            valid = True

        return valid

    def execute(self, context):
        nodes = context.space_data.node_tree.nodes
        
        for n in nodes:
            if self.direction == 'LEFT':                
                n.select = n.location.x < self.location
            else:
                n.select = n.location.x > self.location

        return {'FINISHED'}
    
    def invoke(self, context, event):        
        self.location = context.space_data.cursor_location.x
        return self.execute(context)

classes.append(DJH_OT_NodesSelectByLoc)


class DJH_MT_UtilsMenuMesh(bpy.types.Menu):
    bl_idname = 'DJH_MT_utils_mesh'
    bl_label = 'DJH Utils'

    def draw(self, context):
        l = self.layout

        l.operator('djh.snap_to_faces', text='Snap to Near Faces')
        l.operator('djh.knife_project_sel', text="Knife Project Selected")

classes.append(DJH_MT_UtilsMenuMesh)

class DJH_MT_UtilsMenuObject(bpy.types.Menu):
    bl_idname = 'DJH_MT_utils_object'
    bl_label = 'DJH Utils'

    def draw(self, context):
        l = self.layout
        
        l.operator('djh.clean_vert_groups', text='Clean Vertex Groups')
        l.operator('djh.rename_modifiers', text='Auto-Name Modifiers')
        l.operator('djh.placeholder_materials', text='Create Materials for Instances')
        l.operator('djh.randomize_seeds', text='Randomize Seeds')
        l.operator('djh.freeze_modifier', text='Freeze Modifier into Mesh')
        l.operator('djh.modifier_apply_above', text='Apply Modifiers Above')

classes.append(DJH_MT_UtilsMenuObject)

class DJH_MT_UtilsMenuNodes(bpy.types.Menu):
    bl_idname = 'DJH_MT_utils_node'
    bl_label = 'DJH Utils'

    def draw(self, context):
        l = self.layout
        
        l.operator('djh.get_textures', text='Get Material Textures')
        l.operator('djh.bridge_nodes', text='Bridge Node Sockets (by name)')
        l.operator('djh.select_nodes_by_loc', text='Select to Left').direction = 'LEFT'
        l.operator('djh.select_nodes_by_loc', text='Select to Right').direction = 'RIGHT'

classes.append(DJH_MT_UtilsMenuNodes)

def draw_uv_menu_item(self, context):
        layout = self.layout
        layout.operator('uv.unwrap',text="Unwrap (Fast)").method = 'ANGLE_BASED'

def register():
    def draw_utils_menu_mesh(self, context):
        layout = self.layout
        layout.menu("DJH_MT_utils_mesh")

    def draw_utils_menu_obj(self, context):
        layout = self.layout
        layout.menu("DJH_MT_utils_object")

    def draw_utils_menu_node(self, context):
        layout = self.layout
        layout.menu("DJH_MT_utils_node")


    for c in classes:
        bpy.utils.register_class(c)

    bpy.types.VIEW3D_MT_edit_mesh.append(draw_utils_menu_mesh)
    bpy.types.VIEW3D_MT_object.append(draw_utils_menu_obj)
    bpy.types.NODE_MT_node.append(draw_utils_menu_node)
    bpy.types.VIEW3D_MT_uv_map.prepend(draw_uv_menu_item)
    bpy.types.IMAGE_MT_uvs_unwrap.prepend(draw_uv_menu_item)

def unregister():
    for c in classes:
        bpy.utils.unregister_class(c)

if __name__ == '__main__':
    register()
